"""
msgzip.py
(c) by Michael Stroeder <michael@stroeder.com>

This module is distributed under the terms of the
GPL (GNU GENERAL PUBLIC LICENSE) Version 2
(see http://www.gnu.org/copyleft/gpl.html)

$Id: msgzip.py,v 1.7 2002/06/30 15:47:21 michael Exp $
"""

try:

  import gzip

except ImportError:
  GzipFile=None

else:
  class GzipFile(gzip.GzipFile):
    """
    Stub class for gzip.GzipFile with delayed output of gzip-header
    """
    def __init__(self,filename=None,mode=None,compresslevel=9,fileobj=None):
      gzip.GzipFile.__init__(
        self,filename,mode,compresslevel,fileobj
      )
      self.compresslevel = compresslevel
      self._initialize = 1

    def write(self,data):
      if self._initialize:
        if self.mode == gzip.WRITE:
            self._write_gzip_header()
	self._initialize = 0
      gzip.GzipFile.write(self,data)

  class NewGzipFile(gzip.GzipFile):
    """
    Stub class for gzip.GzipFile with delayed output of gzip-header
    """
    def __defer_init__(self,filename=None,mode=None,compresslevel=9,fileobj=None):
      self._defer_init_filename = filename
      self._defer_init_mode = mode
      self.compresslevel = compresslevel
      self._defer_init_fileobj = fileobj
      self.write = self.first_write_with_header

    def first_write_with_header(self,data):
      # Do a deferred __init__()
      gzip.GzipFile.__init__(
        self,
        self._defer_init_filename,
        self._defer_init_mode,
        self.compresslevel,
        self._defer_init_fileobj
      )
      self.write = self._write
      gzip.GzipFile.write(self,data)

    def _write(self,data):
      gzip.GzipFile.write(self,data)

