"""
w2lapp.schema.viewer -  Display LDAPv3 schema

web2ldap - a web-based LDAP Client,
see http://www.web2ldap.de for details

(c) by Michael Stroeder <michael@stroeder.com>

This module is distributed under the terms of the
GPL (GNU GENERAL PUBLIC LICENSE) Version 2
(see http://www.gnu.org/copyleft/gpl.html)

$Id: viewer.py,v 1.31 2002/11/14 18:24:00 michael Exp $
"""

import types,ldap.schema,w2lapp.gui

from w2lapp.core import utf2display


def LinkText(se):
  names = se.__dict__.get('names',[])
  if len(names)==1:
    link_text = names[0]
  elif len(names)>1:
    link_text = '%s (alias %s)' % (names[0],', '.join(names[1:]))
  else:
    link_text = se.oid
  return link_text


def displayNameOrOID(sid,outf,form,ls,dn,schema,se_nameoroid,se_class):
  se = schema.get_obj(se_class,se_nameoroid,default=None)
  if se:
    link_text = LinkText(se)
    return w2lapp.gui.W2L_Anchor(
      form,'oid',link_text,sid,[
        ('dn',dn),('oid',se.oid)
      ]
    )
  else:
    return se_nameoroid


def displayNameOrOIDList(sid,outf,form,ls,dn,schema,se_names,se_class):
  link_texts = []
  for se_nameoroid in se_names:
    se = schema.get_obj(se_class,se_nameoroid,default=None)
    if se:
      link_text = LinkText(se)
      anchor = w2lapp.gui.W2L_Anchor(
        form,'oid',link_text,sid,[
          ('dn',dn),('oid',se.oid)
        ]
      )
      link_texts.append((link_text,anchor))
    else:
      link_texts.append((se_nameoroid,se_nameoroid))
  link_texts.sort()
  return [ i[1] for i in link_texts ]


def HTMLSchemaTree(sid,outf,form,ls,dn,schema,se_class,se_tree,se_oid,level):
  """HTML output for browser"""
  outf.write('<dl>')
  se_obj = schema.get_obj(se_class,se_oid)
  if se_obj!=None:
    if se_obj.names:
      display_id = se_obj.names[0]
    else:
      display_id = se_oid
    outf.write("""
    <dt><strong>%s</strong></dt>
    """ % (displayNameOrOID(sid,outf,form,ls,dn,schema,display_id,se_class)))
  if se_tree[se_oid]:
    outf.write('<dd>')
    for sub_se_oid in se_tree[se_oid]:
      HTMLSchemaTree(sid,outf,form,ls,dn,schema,se_class,se_tree,sub_se_oid,level+1)
    outf.write('</dd>')
  outf.write('</dl>')


def SchemaContextMenu(sid,form,ls,dn):
  """Build context menu with schema-related items"""
  context_menu_list=[
    w2lapp.gui.W2L_Anchor(
      form,'oid','Browse Schema',sid,[('dn',dn)]
    ),
  ]
  try:
    subschemaSubentryDN = ls.searchSubSchemaEntryDN(dn)
  except ldap.LDAPError:
    pass
  else:
    if subschemaSubentryDN!=None:
      form_param_list = [
        ('dn',subschemaSubentryDN),
      ]
      for schema_attr in ldap.schema.SCHEMA_ATTRS+['objectClass','cn']:
        form_param_list.append(('read_attr',schema_attr))
      context_menu_list.append(w2lapp.gui.W2L_Anchor(
        form,'read','Subschema sub entry',sid,form_param_list
      ))
  return context_menu_list


class DisplaySchemaElement:
  detail_attrs = []

  def __init__(self,sub_schema,se):
    self.s = sub_schema
    self.se = se

  def displayDetails(self,sid,outf,form,ls,dn):
    for text,class_attr,se_class in self.detail_attrs:
      class_attr_value = self.se.__dict__.get(class_attr,None)
      if not class_attr_value:
        continue
      if se_class is None:
        value_output = class_attr_value
      elif type(class_attr_value)==types.ListType:
        class_attr_value.sort()
        value_output = ', '.join(displayNameOrOIDList(sid,outf,form,ls,dn,self.s,class_attr_value,se_class))
      elif type(class_attr_value)==types.StringType:
        value_output = displayNameOrOID(sid,outf,form,ls,dn,self.s,class_attr_value,se_class)
      else:
        value_output = repr(class_attr_value)
      outf.write('<dt>%s</dt>\n<dd><p>\n%s\n</p></dd>\n' % (text,value_output))
    return

  def display(self,sid,outf,form,ls,dn):
    w2lapp.gui.TopSection(
      sid,outf,form,ls,dn,
      '%s %s (%s)' % (
        self.type_desc,
        ', '.join(
          self.se.__dict__.get('names',[])
        ),
        self.se.oid
      ),
      w2lapp.gui.MainMenu(sid,form,ls,dn),
      context_menu_list=SchemaContextMenu(sid,form,ls,dn)
    )
    outf.write("""<div id="MessageDiv">
<p>%s</p>
<h1>%s <em>%s</em> (%s)</h1>
<dl>
<dt>Schema element string:</dt>
<dd><code>%s</code></dd>
</dl>
<p>
    """ % (
    OIDInputForm(form,sid,''),
    self.type_desc,
    ", ".join(
      self.se.__dict__.get('names',[])
    ),
    self.se.oid,
    utf2display(form.accept_charset,str(self.se))
    ))
    self.displayDetails(sid,outf,form,ls,dn)
    outf.write('</p></div>')
    w2lapp.gui.PrintFooter(outf,form)


class DisplayObjectClass(DisplaySchemaElement):
  type_desc = 'Object class'
  detail_attrs = [
    ('Description','desc',None),
    ('Derived from','sup',ldap.schema.ObjectClass),
#    ('Extends required attributes','must',ldap.schema.AttributeType),
#    ('Extends allowed attributes','may',ldap.schema.AttributeType),
  ]

  def displayDetails(self,sid,outf,form,ls,dn):
    DisplaySchemaElement.displayDetails(self,sid,outf,form,ls,dn)
    must,may = self.s.attribute_types([self.se.oid],raise_keyerror=0)
    # Display all required and allowed attributes
    outf.write('<dt>All required attributes:</dt><dd>\n%s</dd>\n' % (
      ', '.join(displayNameOrOIDList(sid,outf,form,ls,dn,self.s,must.keys(),ldap.schema.AttributeType))
    ))
    outf.write('<dt>All allowed attributes:</dt><dd>\n%s</dd>\n' % (
      ', '.join(displayNameOrOIDList(sid,outf,form,ls,dn,self.s,may.keys(),ldap.schema.AttributeType))
    ))
    # Display tree of derived attribute types
    outf.write('<dt>Object class tree:</dt>\n')
    outf.write('<dd>\n')
    try:
      oc_tree = self.s.tree(ldap.schema.ObjectClass)
    except KeyError,e:
      outf.write('<strong>Missing schema elements referenced:<pre>%s</pre></strong>\n' % utf2display(form.accept_charset,str(e)))
    else:
      if oc_tree.has_key(self.se.oid) and oc_tree[self.se.oid]:
        HTMLSchemaTree(sid,outf,form,ls,dn,self.s,ldap.schema.ObjectClass,oc_tree,self.se.oid,0)
    outf.write('</dd>\n')
    

class DisplayAttributeType(DisplaySchemaElement):
  type_desc = 'Attribute type'
  detail_attrs = [
    ('Description','desc',None),
    ('Derived from','sup',ldap.schema.AttributeType),
    ('LDAP Syntax','syntax',ldap.schema.LDAPSyntax),
    ('Equality matching rule','equality',ldap.schema.MatchingRule),
    ('Sub-string matching rule','substr',ldap.schema.MatchingRule),
    ('Ordering matching rule','ordering',ldap.schema.MatchingRule),
  ]

  def displayDetails(self,sid,outf,form,ls,dn):
    DisplaySchemaElement.displayDetails(self,sid,outf,form,ls,dn)
    # Display list of object classes which allow this attribute type
    all_object_classes = self.s.listall(ldap.schema.ObjectClass)
    attr_type_usable_in = []
    for oc_oid in all_object_classes:
      must,may = self.s.attribute_types([oc_oid],raise_keyerror=0)
      if must.has_key(self.se.oid) or may.has_key(self.se.oid):
        attr_type_usable_in.append(oc_oid)
    if attr_type_usable_in:
      outf.write('<dt>Attribute type usable in:</dt>\n<dd>\n%s\n</dd>\n' % (
        ', '.join(displayNameOrOIDList(sid,outf,form,ls,dn,self.s,attr_type_usable_in,ldap.schema.ObjectClass))
      ))
    outf.write('<dt>Attribute type tree:</dt>\n')
    outf.write('<dd>\n')
    # Display tree of derived attribute types
    try:
      at_tree = self.s.tree(ldap.schema.AttributeType)
    except KeyError,e:
      outf.write('<strong>Missing schema elements referenced:<pre>%s</pre></strong>\n' % utf2display(form.accept_charset,str(e)))
    else:
      if at_tree.has_key(self.se.oid) and at_tree[self.se.oid]:
        HTMLSchemaTree(sid,outf,form,ls,dn,self.s,ldap.schema.AttributeType,at_tree,self.se.oid,0)
    outf.write('</dd>\n')
    

class DisplayLDAPSyntax(DisplaySchemaElement):
  type_desc = 'LDAP Syntax'
  detail_attrs = [
    ('Description','desc',None),
  ]

  def displayDetails(self,sid,outf,form,ls,dn):
    DisplaySchemaElement.displayDetails(self,sid,outf,form,ls,dn)
    # Display list of attribute types which directly reference this syntax
    syntax_used_for = [
      at_oid
      for at_oid in self.s.listall(ldap.schema.AttributeType)
      if self.s[at_oid].__class__==ldap.schema.AttributeType and \
         self.s.get_syntax(at_oid)==self.se.oid
    ]
    if syntax_used_for:
      outf.write('<dt>Syntax used for:</dt>\n<dd>\n%s\n</dd>\n' % (
        ', '.join(displayNameOrOIDList(sid,outf,form,ls,dn,self.s,syntax_used_for,ldap.schema.AttributeType))
      ))


class DisplayMatchingRule(DisplaySchemaElement):
  type_desc = 'Matching Rule'
  detail_attrs = [
    ('Description','desc',None),
    ('LDAP syntax','syntax',ldap.schema.LDAPSyntax),
    ('Applies to','applies',ldap.schema.AttributeType),
  ]

SCHEMA_VIEWER_CLASS = {
  ldap.schema.ObjectClass:DisplayObjectClass,
  ldap.schema.AttributeType:DisplayAttributeType,
  ldap.schema.LDAPSyntax:DisplayLDAPSyntax,
  ldap.schema.MatchingRule:DisplayMatchingRule,
}


def OIDInputForm(form,sid,oid=''):
  return '<div id="InputDiv">\n<p>\n%s\n</p></div>\n' % w2lapp.gui.W2L_Form(
    form,'oid','Search',sid,[],
    extrastr=w2lapp.gui.OIDInput('oid',u'OID').inputHTML(oid)
  )


def DisplayAllSchemaelements(sid,outf,form,ls,dn,schema,oid_list=None):
  oid_list = oid_list or []
  w2lapp.gui.TopSection(
    sid,outf,form,ls,dn,'Schema elements',
    w2lapp.gui.MainMenu(sid,form,ls,dn),
    context_menu_list=SchemaContextMenu(sid,form,ls,dn)
  )
  outf.write(OIDInputForm(form,sid,''))
  oid_dict = {}
  if oid_list:
    for schema_class in SCHEMA_VIEWER_CLASS.keys():
      oid_dict[schema_class] = []
    for oid in oid_list:
      oid_dict[schema[oid].__class__].append(oid)
  else:
    for schema_class in SCHEMA_VIEWER_CLASS.keys():
      oid_dict[schema_class] = schema.listall(schema_class)
  outf.write('<div id="MessageDiv">\n<h1>Schema elements</h1>\n')
  for schema_class in SCHEMA_VIEWER_CLASS.keys():
    schema_elements = oid_dict[schema_class]
    if not schema_elements:
      continue
    outf.write('<h2>%ss</h2>\n' % (SCHEMA_VIEWER_CLASS[schema_class].type_desc))
    outf.write(
      '<p>found %d %ss</p>\n' % (
        len(schema_elements),SCHEMA_VIEWER_CLASS[schema_class].type_desc.lower()
    ))
    outf.write(
      ', '.join(
        displayNameOrOIDList(
          sid,outf,form,ls,dn,schema,schema_elements,schema_class
        )
      )
    )    
  outf.write('</div>\n')
  w2lapp.gui.PrintFooter(outf,form)


def w2l_DisplaySchemaElement(sid,outf,command,form,ls,dn):

  sub_schema = ls.retrieveSubSchema(dn,w2lapp.cnf.GetParam(ls,'_schema',None))

  # Get input parameter from form input
  oid = form.getInputValue('oid',[None])[0]

  if oid is None:
    # Display entry page of schema browser
    DisplayAllSchemaelements(sid,outf,form,ls,dn,sub_schema)
  else:
    # Display single schema element specified by OID
    if not sub_schema.has_key(oid):
      oid_list = []
      for schema_element_type in SCHEMA_VIEWER_CLASS.keys():
        if sub_schema.name2oid.get(schema_element_type,{}).has_key(oid):
          oid_list.append(sub_schema.name2oid[schema_element_type][oid])
      if not oid_list:
        raise w2lapp.core.ErrorExit('Name or OID not found in schema!%s' % OIDInputForm(form,sid,oid))
      elif len(oid_list)==1:
        oid = oid_list[0]
      else:
        DisplayAllSchemaelements(sid,outf,form,ls,dn,sub_schema,oid_list)

    se = sub_schema[oid]
    if not SCHEMA_VIEWER_CLASS.has_key(se.__class__):
      raise w2lapp.core.ErrorExit('No viewer for this type of schema element!')

    schema_viewer = SCHEMA_VIEWER_CLASS[se.__class__](sub_schema,se)
    schema_viewer.display(sid,outf,form,ls,dn)
